/*
 * struct.c  Copyright (C) 1997,1998 Stan Brooks
 * SLang
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include <slang.h>

typedef void _SLang_Struct_Type;
typedef void _SLstruct_Field_Type;

extern int _SLstruct_get_field(char *);
extern int _SLang_push_null();
extern int _SLang_pop_struct(_SLang_Struct_Type **);
extern int _SLang_push_struct(_SLang_Struct_Type * s);
extern _SLstruct_Field_Type *pop_field(_SLang_Struct_Type * s, char *name);

extern int init_Utils(void);

extern int init_SLtcp(void);
extern int SLtcp_open(int *, const char *);

// Just an idle thought:
// p = s_find(struct p0,struct p,next-field,value-field,cookie)

// p = s_deref(struct,field,n)
static void s_deref_n(char *field, int *rep)
{
	int n = *rep;

	for (; n > 0; n--)
		if (-1 == _SLstruct_get_field(field)) {
			(void) _SLang_push_null();
			break;
		}
	return;
}

// our exit intrinsic:
static void c_exit(int *code)
{
	exit(*code);
}

// because we don't like the "\r\n" used in slang itself:
static void Local_vmessage(char *fmt, va_list ap)
{
	vfprintf(stdout, fmt, ap);
	fputs("\n", stdout);
}

#define MAX_FILE_LINE_LEN 256

typedef struct
{
   char *buf;
   FILE *fp;
}
File_Client_Data_Type;

static char *read_from_file (SLang_Load_Type *x)
{
   File_Client_Data_Type *c;

   c = (File_Client_Data_Type *)x->client_data;
   return fgets (c->buf, MAX_FILE_LINE_LEN, c->fp);
}

/* Note that file could be freed from Slang during run of this routine
 * so get it and store it !! (e.g., autoloading)
 */

static int Local_SLang_load_file(char *f)
{
	File_Client_Data_Type client_data;
	SLang_Load_Type *x;
	char *name, *buf;
	FILE *fp;

	name = SLang_create_slstring(f);
	if (name == NULL) return -1;

	if (NULL == (x = SLallocate_load_type(name))) {
		SLang_free_slstring(name);
		return -1;
	}
	buf = NULL;

	fp = fopen(f, "r");
	if (fp == NULL)
		SLang_verror(SL_OBJ_NOPEN, "Unable to open %s", name);
	else if (NULL != (buf = SLmalloc(MAX_FILE_LINE_LEN + 1))) {
		client_data.fp = fp;
		client_data.buf = buf;
		x->client_data = (VOID_STAR) & client_data;
		x->read = read_from_file;

		(void) SLang_load_object(x);
	}
	fclose(fp);

	SLfree(buf);
	SLang_free_slstring(name);
	SLdeallocate_load_type(x);

	if (SLang_Error) return -1;

	return 0;
}

/* Create the Table that S-Lang requires */
static SLang_Intrin_Fun_Type TTEST_Intrinsics[] =
{
	MAKE_INTRINSIC_I("exit", c_exit, VOID_TYPE),
	MAKE_INTRINSIC_SI("s_deref_n", s_deref_n, VOID_TYPE),
	SLANG_END_TABLE
};

int main(int argc, char **argv)
{

	char *file;

	/* parse command line arguments */
	if (argc != 2) {
		fprintf(stderr, "Usage: %s FILENAME\n", argv[0]);
		exit(-1);
	}
	file = argv[1];

	/* Initialize the library.  This is always needed. */

	if (SLang_init_slang()				/* basic interpreter functions */
			||SLang_init_slmath()			/* sin, cos, etc... */
			||SLang_init_slunix()			/* unix system calls */
			||SLang_init_slfile()			/* file i/o */
			) {
		fprintf(stderr, "Unable to initialize S-Lang.\n");
		exit(-1);
	}
	SLang_VMessage_Hook = Local_vmessage;
	SLang_Load_File_Hook = Local_SLang_load_file;

	/* Now add intrinsics for this application */
	if (SLadd_intrin_fun_table(TTEST_Intrinsics, NULL)) {
		fprintf(stderr, "Unable to add intrinsics.\n");
		exit(-1);
	}
	if (!init_Utils()) {
		fprintf(stderr, "Ooops. Utils Init failed.\n");
		exit(-1);
	}
	if (!init_SLtcp()) {
		fprintf(stderr, "Ooops. SLtcp Init failed.\n");
		exit(-1);
	}
	// Turn on debugging
	SLang_Traceback = 1;

	/* Now load the slang-script */
	SLang_load_file(file);

	return (SLang_Error);
}
